package de.osci.osci12.samples.impl.crypto;

import java.io.IOException;
import java.io.InputStream;
import java.security.Key;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.UnrecoverableKeyException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Enumeration;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.encryption.OSCICipherException;


/**
 * Diese Klasse ist eine Beispiel-Implementierung der abstrakten Decrypter-Klasse.
 * Fr die Verwendung wird ein PKCS#12-Keystore in Form einer *.p12-Datei bentigt.
 * Die Implementierung ist fr Testzwecke bestimmt, sie greift auf den ersten
 * verfgbaren Alias zu. Die PIN fr dessen Privatschlssel muss die gleiche sein
 * wie die des Keystores.
 *
 * Diese einfache Implementierung hlt die PIN des Keystores als Character-Array
 * im Arbeitsspeicher, sie wird als String bergeben. Es wird Anwendern empfohlen,
 * eigene Implementierungen zu schreiben, die die PIN in der Methode decrypt(...)
 * abfragen und nach Gebrauch wieder lschen oder anderweitig fr ein sicheres
 * Pin-Cashing zu sorgen.
 *
 * <p>Copyright  2004 Freie Hansestadt Bremen</p>
 * <p>Erstellt von bremen online services Entwicklungs- und
 * Betriebsgesellschaft GmbH & Co. KG</p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL www.osci.de abgerufen oder bei der OSCI-Leitstelle,
 * Senator fr Finanzen, Freie Hansestadt Bremen, Postfach 10 15 40,
 * 28015 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.2.3
 * @see de.osci.osci12.extinterfaces.crypto.Decrypter
 */
public class PKCS12Decrypter extends de.osci.osci12.extinterfaces.crypto.Decrypter
{
  private static Log log = LogFactory.getLog(PKCS12Decrypter.class);
  private X509Certificate cert;
  private String alias;
  private char[] pin = null;
  private KeyStore keyStore = null;
  private Key key;

  /**
   * Legt ein neues PKCS12Decrypter Objekt an.
   *
   * @param p12_fileName Pfad zur PKSC#12-Keystore Datei
   * @param pin PIN
   *
   * @throws KeyStoreException wenn Probleme beim Laden des Keystores auftreten
   * @throws CertificateException bei Problemen bei der Verarbeitung der Zertifikate
   * @throws NoSuchAlgorithmException falls nicht untersttzte Algotrihmen erforderlich sind
   * @throws IOException bei Schreib-/Leseproblemen
   * @throws UnrecoverableKeyException Fehler beim Einlesen der Schlssel
   */
  public PKCS12Decrypter(String p12_fileName, String pin)
                  throws KeyStoreException,
                         CertificateException,
                         NoSuchAlgorithmException,
                         IOException,
                         UnrecoverableKeyException
  {
    init(PKCS12Decrypter.class.getResourceAsStream(p12_fileName), pin);
  }

  /**
   * Legt ein neues PKCS12Decrypter Objekt an.
   *
   * @param in InputStream des PKSC#12-Keystores
   * @param pin PIN
   *
   * @throws KeyStoreException wenn Probleme beim Laden des Keystores auftreten
   * @throws CertificateException bei Problemen bei der Verarbeitung der Zertifikate
   * @throws NoSuchAlgorithmException falls nicht untersttzte Algotrihmen erforderlich sind
   * @throws IOException bei Schreib-/Leseproblemen
   * @throws UnrecoverableKeyException Fehler beim Einlesen der Schlssel
   */
  public PKCS12Decrypter(InputStream in, String pin)
                  throws KeyStoreException,
                         CertificateException,
                         NoSuchAlgorithmException,
                         IOException,
                         UnrecoverableKeyException
  {
    init(in, pin);
  }

  private void init(InputStream in, String pin)
             throws KeyStoreException,
                    CertificateException,
                    NoSuchAlgorithmException,
                    IOException,
                    UnrecoverableKeyException
  {
    this.pin = pin.toCharArray();

    if (DialogHandler.getSecurityProvider() == null)
      keyStore = java.security.KeyStore.getInstance("PKCS12");
    else
      keyStore = java.security.KeyStore.getInstance("PKCS12", DialogHandler.getSecurityProvider());
   
    keyStore.load(in, this.pin);

    String al = null;
    Enumeration<String> e = keyStore.aliases();

    while (e.hasMoreElements())

      if (keyStore.isKeyEntry(al = e.nextElement()))
        break;

    if (al == null)
      throw new NullPointerException("No private key found in keystore.");

    this.alias = al;
    cert = (X509Certificate) keyStore.getCertificate(alias);
    key = keyStore.getKey(alias, this.pin);
  }

  /**
   * Liefert den Namen des Herstellers.
   * @return Herstellername
   */
  public String getVendor()
  {
    return "BOS";
  }

  /**
   * Liefert die Versionsnummer.
   * @return Versionsnummer
   */
  public String getVersion()
  {
    return "1.6";
  }

  /**
   *  Liefert das Verschlsselungszertifikat.
   *
   *@return  das Verschlsselungszertifikat
   */
  public X509Certificate getCertificate()
  {
    return cert;
  }

  /**
   *  Entschlsselt das bergebene Byte-Array gem RSAES-PKCS1-v1_5
   *
   *@param  data die zu entschlsselnden Daten
   *@return verschlsselten Daten
   */
  public byte[] decrypt(byte[] data) throws de.osci.osci12.encryption.OSCICipherException
  {
    try
    {
      return de.osci.osci12.encryption.Crypto.doRSADecryption(key, data);
    }
    catch (Exception ex)
    {
      throw new de.osci.osci12.encryption.OSCICipherException("decryption_error");
    }
  }

  /**
   *  Entschlsselt das bergebene Byte-Array gem RSAES-OAEP
   *
   *@param  data die zu entschlsselnden Daten
   *@return verschlsselten Daten
   */
  public byte[] decrypt(byte[] data, String mgfAlgorithm, String digestAlgorithm) throws OSCICipherException,
                                                                    de.osci.osci12.common.OSCICancelledException
  {
    try
    {
      return de.osci.osci12.encryption.Crypto.doRSADecryption(key, data, Constants.ASYMMETRIC_CIPHER_ALGORITHM_RSA_OAEP, mgfAlgorithm, digestAlgorithm, null);
    }
    catch (Exception ex)
    {
      throw new de.osci.osci12.encryption.OSCICipherException("decryption_error");
    }
  }

}
