package de.osci.helper;

import java.io.FilterInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.GeneralSecurityException;
import java.security.NoSuchAlgorithmException;

import javax.crypto.BadPaddingException;
import javax.crypto.Cipher;
import javax.crypto.CipherInputStream;
import javax.crypto.SecretKey;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.encryption.Crypto;


/**
 * Diese Erweiterung der Klasse java.io.FilterInputStream fhrt eine symmetrische Ver-/Entschlsselung der
 * gelesenen Daten durch. Der Initialisierungsvektor wird gem der XML-Encryption-Spezifikation den
 * Inhaltsdaten hinzugefgt bzw. entnommen.
 * <p>
 * Copyright  2017 Freie Hansestadt Bremen
 * </p>
 * <p>
 * Erstellt von Governikus GmbH & Co.  KG
 * </p>
 * <p>
 * Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr freie Softwarebibliotheken genutzt
 * werden.
 * </p>
 * Die Lizenzbestimmungen knnen unter der URL <a href='www.osci.de' >www.osci.de</a > abgerufen oder bei der Senator fr
 * Finanzen, Freie Hansestadt Bremen,  Schillerstr. 1, 28195 Bremen schriftlich angefordert werden.
 * </p>
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.8.0
 * @see de.osci.helper.SymCipherOutputStream
 */
public class SymCipherInputStream extends FilterInputStream
{

  private static Log log = LogFactory.getLog(SymCipherInputStream.class);
  private Cipher cipher;
  private String symCipherAlgo=null;
  private int count = 0;

  byte[] iv;

  private boolean encrypt;

  private byte[] b = new byte[1];

  /**
   * Creates a new SymCipherInputStream object, use default algorithm for symmetric key (see
   * {@link Crypto.getCipherAlgoID}).
   *
   * @param inStream undocumented
   * @param symKey undocumented
   * @param encrypt undocumented
   * @throws IOException undocumented
   * @throws NoSuchAlgorithmException
   * @throws IllegalArgumentException undocumented
   */
  public SymCipherInputStream(InputStream inStream, String symAlgorithm, boolean encrypt)
    throws IOException, NoSuchAlgorithmException
  {
    this(inStream, Crypto.createSymKey(symAlgorithm), symAlgorithm, encrypt);
  }

  /**
   * Creates a new SymCipherInputStream object with given algorithm.
   *
   * @param inStream undocumented
   * @param symKey undocumented
   * @param encrypt undocumented
   * @throws IOException undocumented
   * @throws IllegalArgumentException undocumented
   */
  public SymCipherInputStream(InputStream inStream, SecretKey symKey, String symAlgorithm, boolean encrypt)
    throws IOException
  {
    super(null);
    symCipherAlgo=symAlgorithm;
    this.encrypt = encrypt;

    if (inStream == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " null");

    if (symKey == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_secondargument") + " null");
    try
    {
      if (symAlgorithm == null)
      {
        throw new IllegalArgumentException(DialogHandler.text.getString("invalid_thirdargument") + " null");
        // symAlgorithm = Crypto.getCipherAlgoID(symKey);
      }
      if(ParserHelper.isGCMAlgorithmOnly()&& symAlgorithm.endsWith("-cbc"))
      {
        throw new IllegalArgumentException(DialogHandler.text.getString("invalid_thirdargument") + " Wrong algorithm! only GCM supported! Given: "+symAlgorithm);
      }
      if (DialogHandler.getSecurityProvider() == null)
        cipher = Cipher.getInstance(Constants.JCA_JCE_MAP.get(symAlgorithm));
      else
        cipher = Cipher.getInstance(Constants.JCA_JCE_MAP.get(symAlgorithm),
                                    DialogHandler.getSecurityProvider());

      if (encrypt)
      {
        cipher.init(Cipher.ENCRYPT_MODE, symKey);
        iv = cipher.getIV();
      }
      else
      {
        if (symKey.getAlgorithm().startsWith("DESede"))
          iv = new byte[8];
        else
          iv = new byte[16];

        int tmp;

        for ( int i = 0 ; i < iv.length ; i++ )
        {
          tmp = inStream.read();

          if (tmp == -1)
            throw new IOException("Unexpected EOS");

          iv[i] = (byte)tmp;
        }

        javax.crypto.spec.IvParameterSpec algoParamSpec = new javax.crypto.spec.IvParameterSpec(iv);
        cipher.init(Cipher.DECRYPT_MODE, symKey, algoParamSpec);
      }
    }
    catch (GeneralSecurityException ex)
    {
      if (encrypt)
        throw new IOException(DialogHandler.text.getString("encryption_error"));
      else
        throw new IOException(DialogHandler.text.getString("decryption_error"));
    }

    super.in = new CipherInputStream(inStream, cipher);
  }

  /**
   * undocumented
   *
   * @return undocumented
   * @throws IOException undocumented
   */
  @Override
  public int read() throws IOException
  {
    if (read(b, 0, 1) == -1)
      return -1;

    return (b[0]) & 0xff;
  }

  /**
   * undocumented
   *
   * @param b undocumented
   * @param off undocumented
   * @param len undocumented
   * @return undocumented
   * @throws IOException undocumented
   */
  @Override
  public int read(byte[] b, int off, int len) throws IOException
  {
    if (encrypt && (count < iv.length))
    {
      if (len < (iv.length - count))
      {
        System.arraycopy(iv, count, b, off, len);
        count += len;

        return len;
      }
      else
      {
        System.arraycopy(iv, count, b, off, iv.length - count);

        try
        {
          int ret = in.read(b, (off + iv.length) - count, len - iv.length + count);
          ret += (iv.length - count);
          count = iv.length;
          return ret;
        }
        catch (IOException ex)
        {
          if (findException(ex, BadPaddingException.class))
          {
            log.error("Bad padding exception. Return end of stream!!!",ex);
            return -1;
          }else
          {
            throw ex;
          }
       }
        
      }
    }
    else
    {
      try
      {

        return in.read(b, off, len);
      }
      catch (IOException ex)
      {
        if (findException(ex, BadPaddingException.class))
        {
          log.error("Bad padding exception. Return end of stream!!!",ex);
          return -1;
        }else
        {
          throw ex;
        }
      }
    }
  }

  private boolean findException(Throwable ex, Class<? extends Throwable> findEx)
  {
    // nur bei CBC kann es wirklich eine BadPadding Exception geben
    if (symCipherAlgo.endsWith("-gcm"))
    {
      return false;
    }
    while (ex.getCause() != null)
    {
      if (findEx.isInstance(ex.getCause()))
        return true;
      else
        ex = ex.getCause();
    }
    return false;
  }
}
