package de.osci.osci12.encryption;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.security.DigestOutputStream;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.ContentContainer;


/**
 * <p>Copyright  2017 Freie Hansestadt Bremen</p>
 * <p>Erstellt von Governikus GmbH & Co.  KG
 * </p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL <a href='www.osci.de' >www.osci.de</a > abgerufen oder bei der
 * Koordinierungsstelle fr IT-Standards (KoSIT),  Schillerstr. 1,
 * 28195 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.8.0
 */
public class EncryptedData extends EncryptedType
{
  private static Log log = LogFactory.getLog(EncryptedData.class);
  private static byte[] ns;
  public String encNS;
  public String soapNSPrefix = "soap";
  public String osciNSPrefix = "osci";
  public String dsNSPrefix = "ds";
  public String xencNSPrefix = "xenc";
  public String xsiNSPrefix = "xsi";

  static
  {
    try
    {
      ns = " xmlns:ds=\"http://www.w3.org/2000/09/xmldsig#\" xmlns:osci=\"http://<a href='www.osci.de' >www.osci.de</a >/2002/04/osci\" xmlns:soap=\"http://schemas.xmlsoap.org/soap/envelope/\" xmlns:xenc=\"http://www.w3.org/2001/04/xmlenc#\" xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"".getBytes(Constants.CHAR_ENCODING);
    }
    catch (UnsupportedEncodingException e)
    {
      // Jede JVM untersttzt UTF-8
    }
  }

  /**
   * Erstellt ein EncryptedData Objekt mit einem CipherValue
   * @param coco zu verschlsselndes ContentContainer-Objekt
   * @param encryptionMethodAlgorithm Identifier des Verschlsselungsalgorithmus.
   * Mgliche Werte sind de.osci.osci12.common.Constants.SYMMETRIC_CIPHER_ALGORITHM_TDES_CBC,
   * Constants.SYMMETRIC_CIPHER_ALGORITHM_AES128/192/256
   * @param iD ID des EncryptedData Objektes
   * @param key symmetrischer Schlssel fr die Nutzdatenverschlsselung
   */
  public EncryptedData(ContentContainer coco, String encryptionMethodAlgorithm, String iD, javax.crypto.SecretKey key)
  {
    if (coco == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " null");

    if (key == null)
      throw new IllegalArgumentException(DialogHandler.text.getString("invalid_firstargument") + " null");

    CipherValue cipherValue = new CipherValue(coco, key,encryptionMethodAlgorithm);
    CipherData cipherData = new CipherData(cipherValue);
    this.setEncryptionMethodAlgorithm(encryptionMethodAlgorithm);
    this.setMimeType("text/xml");
    this.setId(iD);
    setCipherData(cipherData);
  }

  /**
   * Erstellt ein EncryptedData Objekt mit einem CipherReference
   * @param cipherReference Cipher Reference Objekt Nutzdatenverschlsselung
   * @param encryptionMethodAlgorithm Algoritmus
   * @param iD ID des EncryptedData Objektes
   */
  public EncryptedData(CipherReference cipherReference, String encryptionMethodAlgorithm, String iD)
  {
    CipherData cipherData = new CipherData(cipherReference);
    this.setEncryptionMethodAlgorithm(encryptionMethodAlgorithm);
    this.setMimeType("text/xml");
    this.setId(iD);
    setCipherData(cipherData);
  }

  /**
   * Ein Konstruktor fr den Parser
   * @param iD ID des EncryptedData Objektes
   */
  EncryptedData(String iD)
  {
    this.setMimeType("text/xml");
    this.setId(iD);
  }

  /**
   * undocumented
   *
   * @param refId undocumented
   *
   * @return undocumented
   */
  public EncryptedKey findEncrypedKey(String refId)
  {
    EncryptedKey[] keys = this.getKeyInfo().getEncryptedKeys();

    for (int i = 0; i < keys.length; i++)
    {
      if (log.isDebugEnabled())
        log.debug("KeyInfo Eintrag:" + keys[i].getKeyInfo().getRetrievalMethod().getURI());

      if (log.isDebugEnabled())
        log.debug("Vergleiche refID:\n" + refId + "\nmit:\n" + keys[i].getKeyInfo().getRetrievalMethod().getURI() +
                  "\n" + keys[i].getKeyInfo().getRetrievalMethod().getURI().endsWith(refId));

      //id prefix # wird ingnoriert
      if (keys[i].getKeyInfo().getRetrievalMethod().getURI().endsWith(refId))
        return keys[i];
    }

    return null;
  }

  /**
   * undocumented
   *
   * @param out undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIException undocumented
   */
  public void writeXML(OutputStream out) throws IOException,
                                                OSCIException
  {
    writeXML(out, true);
  }

  /**
   * undocumented
   *
   * @param out undocumented
   * @param inner undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIException undocumented
   */
  public void writeXML(OutputStream out, boolean inner)
                throws IOException,
                       OSCIException
  {
    out.write(("<" + xencNSPrefix + ":EncryptedData").getBytes(Constants.CHAR_ENCODING));

    if (!(out instanceof DigestOutputStream) && !inner)
    {
      if (encNS == null)
        out.write(ns);
      else
        out.write(encNS.getBytes(Constants.CHAR_ENCODING));
    }

    if ((getId() != null) && !getId().equals(""))
      out.write((" Id=\"" + this.getId() + "\"").getBytes(Constants.CHAR_ENCODING));

    out.write((" MimeType=\"" + this.getMimeType() + "\"><" + xencNSPrefix + ":EncryptionMethod Algorithm=\"" +
              this.getEncryptionMethodAlgorithm() + "\"></" + xencNSPrefix + ":EncryptionMethod>").getBytes(Constants.CHAR_ENCODING));
    this.getKeyInfo().writeXML(out, dsNSPrefix, xencNSPrefix);
    this.getCipherData().writeXML(out, dsNSPrefix, xencNSPrefix);
    out.write(("</" + xencNSPrefix + ":EncryptedData>").getBytes(Constants.CHAR_ENCODING));
  }
}
