package de.osci.osci12.extinterfaces;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URI;


/**
 * Interface-Klasse fr das bermitteln von OSCI-Nachrichten
 * Die OSCI-Bibliothek sieht nicht vor, den Nutzer auf ein Transportprotokoll
 * festzulegen. Aus diesem Grund wird dieses Transport-Interface zur Verfgung
 * gestellt, welches es der Anwendung ermglicht, die erstellten OSCI-Nachrichten
 * mit dem gewnschtem Protokoll oder auf die gewnschte Art zu bermitteln oder
 * zu speichern. Vorstellbare Implementierungen sind z.B. http, https, ftp,
 * smtp/pop, Filesystem oder jms.
 *
 * <p>Copyright  2017 Freie Hansestadt Bremen</p>
 * <p>Erstellt von Governikus GmbH & Co.  KG
 * </p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL <a href='www.osci.de' >www.osci.de</a > abgerufen oder bei der
 * Koordinierungsstelle fr IT-Standards (KoSIT),  Schillerstr. 1,
 * 28195 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.8.0
 */
public interface TransportI
{
  /**
   * Liefert die Versionsnummer.
   * @return Versionsnummer
   */
  public String getVersion();

  /**
   * Sollte den Namen des Herstellers zurckgeben.
   * @return Herstellername
   */
  public String getVendor();

  /**
   * Die Implementierung dieser statischen Methode muss eine neue Instanz der Klasse
   * zurckgeben.
   * @return neue Instanz der implementierenden Klasse
   * @throws IOException wenn ein Fehler auftritt
   */
  public TransportI newInstance() throws IOException;

  /**
   *  Liefert den Response-Stream.
   *
   *@return den InputStream der eingehenden Antwortdaten
   */
  public InputStream getResponseStream() throws IOException;

  /**
   * Methode kann zur berprfung der Erreichbarkeit einer URL implementiert
   * und verwendet werden. Die Bibliothek selbst ruft diese Methode nicht auf.
   *
   *@param  uri              URI des Kommunikationspartners
   *@return                  true, wenn der Kommunikationspartner erreichbar ist
   *@exception  IOException  im Fehlerfall
   */
  public boolean isOnline(URI uri) throws IOException;

  /**
   *  Sollte die Lnge des Response Streams liefern. Auch diese Methode wird
   * von der Bibliothek z.Zt. nicht bentigt.
   *
   *@return Anzahl der empfangenen Bytes oder <code>-1</code>, falls die Lnge nicht bekannt ist.
   */
  public long getContentLength();

  /**
   *  Liefert eine konkrete Verbindung zum Versenden eines Streams.
   *  Die Methode konnektet zu der bergebenen URI und liefert als Ergebnis
   *  einen Outputstream, in den die Bibliothek dann die serialisierte OSCI-Nachricht
   *  schreibt.
   *
   *@param  uri              URI des Kommunikationspartners
   *@param  length           Lnge der bertragungsdaten (Anz d. Bytes)
   *@return                  Output-Stream, in den die Daten geschrieben werden knnen
   *@exception  IOException  im Fehlerfall
   */
  public OutputStream getConnection(URI uri, long length)
                             throws IOException;
}
