package de.osci.osci12.extinterfaces.crypto;

import java.security.cert.X509Certificate;

import de.osci.osci12.common.Constants;
import de.osci.osci12.encryption.OSCICipherException;


/**
 * Diese abstrakte Klasse stellt die Schnittstelle der Bibliothek fr
 * die Anbindung von Entschlsselungs-Modulen (Crypto-Token) dar. Anwendungen, die
 * OSCI-Nachrichten entschlsseln wollen, mssen an das entschlsselnde Rollenobjekt
 * eine Implementation dieser Klasse bergeben.
 *
 * <p>Copyright  2017 Freie Hansestadt Bremen</p>
 * <p>Erstellt von Governikus GmbH & Co.  KG
 * </p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL <a href='www.osci.de' >www.osci.de</a > abgerufen oder bei der
 * Koordinierungsstelle fr IT-Standards (KoSIT),  Schillerstr. 1,
 * 28195 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N.Bngener
 * @version 1.8.0
 */
public abstract class Decrypter
{
  /**
   * Liefert die Versionsnummer
   * @return Versionsnummer
   */
  public abstract String getVersion();

  /**
   * Sollte den Namen des Herstellers zurckgeben.
   * @return Herstellername
   */
  public abstract String getVendor();

  /**
   * Die Implementierung dieser Methode muss das Verschlsselungszertifikat zurckgeben.
   * @return Verschlsselungszertifikat
   */
  public abstract X509Certificate getCertificate();

  /**
   * Die Implementierung dieser Methode muss das bergebene Byte-Array gem RSAES-PKCS1-v1_5 entschlsseln.
   * @param data zu entschlsselndes Byte-Array
   * @return entschlsseltes Byte-Array
   * @throws OSCICipherException wenn beim Entschlsseln ein Fehler auftritt
   * @throws de.osci.osci12.common.OSCICancelledException wenn der Vorgang vom Anwender abgebrochen wird
   */
  public abstract byte[] decrypt(byte[] data) throws OSCICipherException,
                                                     de.osci.osci12.common.OSCICancelledException;

  /**
   * Die Implementierung dieser Methode muss das bergebene Byte-Array gem RSAES-OAEP-ENCRYPT entschlsseln
   * (P-Source Parameter werden nicht untersttzt).
   * Die Algorithmus-Identifier knnen der Klasse {@link Constants} entnommen werden.
   * <b>Diese Methode muss berschrieben werden, die Implementierung in dieser Klasse wirft eine
   * {@link UnsupportedOperationException}.</b>
   * @param data zu entschlsselndes Byte-Array
   * @param mgfAlgorithm Maskengenerierungsfunktion (OAEP), z.B. http://www.w3.org/2009/xmlenc11#mgf1sha256
   * @param digestAlgorithm Hashalgorithmus (OAEP), z.B. http://www.w3.org/2001/04/xmlenc#sha256
   * @return entschlsseltes Byte-Array
   * @throws OSCICipherException wenn beim Entschlsseln ein Fehler auftritt
   * @throws de.osci.osci12.common.OSCICancelledException wenn der Vorgang vom Anwender abgebrochen wird
   * @see Constants
   * @since 1.6
   */
  public byte[] decrypt(byte[] data, String mgfAlgorithm, String digestAlgorithm) throws OSCICipherException,
                                                     de.osci.osci12.common.OSCICancelledException
  {
    throw new UnsupportedOperationException("RSA-OAEP not implemented yet.");
  }
}
