package de.osci.osci12.messagetypes;

import java.io.IOException;
import java.io.OutputStream;
import java.security.NoSuchAlgorithmException;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants.HeaderTags;
import de.osci.osci12.common.Constants.LanguageTextEntries;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.common.OSCICancelledException;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.roles.Originator;
import de.osci.osci12.soapheader.OsciH;


/**
 * <p>
 * Mit dieser Klasse werden Nachrichtenobjekte fr Zustellungsabholauftrge angelegt. Clients knnen hiermit
 * maximal eine Nachricht vom Intermedir abrufen. Als Antwort auf diese Nachricht erhalten sie vom
 * Intermedir ein ResponseToFetchDelivery-Nachrichtenobjekt, welches eine Rckmeldung ber den Erfolg der
 * Operation (getFeedback()) und ggf. die gewnschte Nachricht enthlt.
 * </p>
 * <p>
 * Copyright  2017 Freie Hansestadt Bremen
 * </p>
 * <p>
 * Erstellt von Governikus GmbH & Co.  KG
 * </p>
 * <p>
 * Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr freie Softwarebibliotheken genutzt
 * werden.
 * </p>
 * Die Lizenzbestimmungen knnen unter der URL <a href='www.osci.de' >www.osci.de</a > abgerufen oder bei der Senator fr
 * Finanzen, Freie Hansestadt Bremen,  Schillerstr. 1, 28195 Bremen schriftlich angefordert werden.
 * </p>
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.8.0
 * @see de.osci.osci12.messagetypes.ResponseToFetchDelivery
 */
public class FetchDelivery extends FetchRequestAbstract
{

  FetchDelivery()
  {
    messageType = FETCH_DELIVERY;
  }

  /**
   * Legt ein Nachrichtenobjekt fr einen Zustellungsabholauftrag an.
   *
   * @param dh DialogHandler-Objekt des Dialogs, innerhalb dessen die Nachricht versendet werden soll
   * @see de.osci.osci12.common.DialogHandler
   */
  public FetchDelivery(DialogHandler dh) throws NoSuchAlgorithmException
  {
    super(dh);
    messageType = FETCH_DELIVERY;
    originator = ((Originator)dh.getClient());
    dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    dialogHandler.getControlblock()
                 .setSequenceNumber(dialogHandler.getControlblock().getSequenceNumber() + 1);
  }


  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck. Diese Methode wirft eine Exception, wenn
   * beim Aufbau oder Versand der Nachricht ein Fehler auftritt. Fehlermeldungen vom Intermedir mssen dem
   * Feedback der Antwortnachricht entnommen werden.
   *
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send(OutputStream, OutputStream)
   */
  public ResponseToFetchDelivery send() throws IOException, OSCIException, NoSuchAlgorithmException
  {
    return (ResponseToFetchDelivery)transmit(null, null);
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck. Die aus- und eingehenden Daten werden
   * zustzlich in die bergebenen Streams geschrieben (unverschlsselte Auftragsdaten). Diese Parameter
   * drfen null sein.
   *
   * @param storeOutput Stream, in den die versendete Nachricht geschrieben wird
   * @param storeInput Stream, in den die empfangene Antwortnachricht geschrieben wird
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send()
   */
  public ResponseToFetchDelivery send(OutputStream storeOutput, OutputStream storeInput)
    throws IOException, OSCIException, NoSuchAlgorithmException
  {
    return (ResponseToFetchDelivery)transmit(storeOutput, storeInput);
  }

  void sign() throws IOException, OSCIException, OSCICancelledException, NoSuchAlgorithmException
  {
    super.sign();
    // Der NonIntemedCertHeader liegt hier weiter unten
    messageParts.set(5, nonIntermediaryCertificatesH);
    messageParts.set(3, null);
  }

  /**
   * undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIException undocumented
   * @throws NoSuchAlgorithmException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void compose() throws IOException, OSCIException, NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(3, null);

    if (dialogHandler.getControlblock().getChallenge() == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name())
                                      + ": Challenge");

    if (dialogHandler.getControlblock().getResponse() == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name())
                                      + ": Response");

    if (dialogHandler.getControlblock().getConversationID() == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name())
                                      + ": Conversation-Id");

    if (dialogHandler.getControlblock().getSequenceNumber() == -1)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name())
                                      + ": SequenceNumber");

    osciH = new OsciH(HeaderTags.fetchDelivery.getElementName(), getSelectionRuleString());
    messageParts.add(osciH);
    messageParts.add(nonIntermediaryCertificatesH);
    if (featureDescription != null && dialogHandler.isSendFeatureDescription())
    {
      messageParts.add(featureDescription);
    }
    messageParts.addAll(customHeaders);
    body = new Body("");
    // body.setNSPrefixes(this);
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
