package de.osci.osci12.messagetypes;

import java.io.IOException;
import java.io.OutputStream;
import java.security.NoSuchAlgorithmException;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.Constants.HeaderTags;
import de.osci.osci12.common.Constants.LanguageTextEntries;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.messageparts.Content;
import de.osci.osci12.messageparts.ContentContainer;
import de.osci.osci12.messageparts.EncryptedDataOSCI;
import de.osci.osci12.roles.Addressee;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.roles.Originator;
import de.osci.osci12.soapheader.OsciH;
import de.osci.osci12.soapheader.QualityOfTimestampH;


/**
 * <p>Mit dieser Klasse werden Nachrichtenobjekte fr Abwicklungsauftrge
 * angelegt. Clients erhalten als Antwort auf diese Nachricht vom Intermedir
 * ein ResponseToMediateDelivery-Nachrichtenobjekt, welches eine Rckmeldung ber den
 * Erfolg der Operation (getFeedback()) und ggf. Inhaltsdaten vom Empfnger enthlt.</p>
 *
 * <p>Copyright  2017 Freie Hansestadt Bremen</p>
 * <p>Erstellt von Governikus GmbH & Co.  KG
 * </p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL <a href='www.osci.de' >www.osci.de</a > abgerufen oder bei der
 * Koordinierungsstelle fr IT-Standards (KoSIT),  Schillerstr. 1,
 * 28195 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.8.0
 *
 * @see de.osci.osci12.messagetypes.ResponseToMediateDelivery
 */
public class MediateDelivery extends OSCIRequest implements ContentPackageI
{
  //  private static Log log = LogFactory.getLog(MediateDelivery.class);
  private String subject;

  /**
   * Legt ein Nachrichtenobjekt fr einen Abwicklungsauftrag an.
   *
   * @param dh DialogHandler-Objekt des Dialogs, innerhalb dessen die Nachricht
   * versendet werden soll
   *
   * @see de.osci.osci12.common.DialogHandler
   */
  public MediateDelivery(DialogHandler dh, Addressee addressee, String uriReceiver)
                  throws java.net.URISyntaxException,
                         NoSuchAlgorithmException
  {
    super(dh);
    messageType = MEDIATE_DELIVERY;
    originator = ((Originator) dh.getClient());
    this.addressee = addressee;
    this.uriReceiver = new java.net.URI(uriReceiver);
    dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
    dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    dialogHandler.getControlblock().setSequenceNumber(dialogHandler.getControlblock().getSequenceNumber() + 1);
    setQualityOfTimeStampCreation(false);
    setQualityOfTimeStampReception(false);
  }

  MediateDelivery()
  {
    super();
    messageType = MEDIATE_DELIVERY;
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Diese Methode wirft eine Exception, wenn beim Aufbau oder Versand der
   * Nachricht ein Fehler auftritt. Fehlermeldungen vom Intermedir mssen
   * dem Feedback der Antwortnachricht entnommen werden.
   *
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send(OutputStream, OutputStream)
   */
  public ResponseToMediateDelivery send() throws IOException,
                                                 de.osci.osci12.OSCIException,
                                                 NoSuchAlgorithmException
  {
    return (ResponseToMediateDelivery) transmit(null, null);
  }

  /**
   * Versendet die Nachricht und liefert die Antwortnachricht zurck.
   * Die aus- und eingehenden Daten werden zustzlich in die bergebenen
   * Streams geschrieben (unverschlsselte Auftragsdaten). Diese Parameter
   * drfen null sein.
   *
   * @param storeOutput Stream, in den die versendete Nachricht geschrieben wird
   * @param storeInput Stream, in den die empfangene Antwortnachricht geschrieben wird
   * @throws IOException bei Ein-/Ausgabefehlern
   * @throws de.osci.osci12.OSCIException bei OSCI-Fehlern
   * @return Antwortnachricht-Objekt
   * @see #send()
   */
  public ResponseToMediateDelivery send(OutputStream storeOutput, OutputStream storeInput)
                                 throws IOException,
                                        de.osci.osci12.OSCIException,
                                        NoSuchAlgorithmException
  {
    return (ResponseToMediateDelivery) transmit(storeOutput, storeInput);
  }

  /**
   * Fgt der Nachricht einen Inhaltsdatencontainer hinzu. Diese Methode
   * sollte erst aufgerufen werden, wenn der Container vollstndig erstellt wurde.
   * <b>Hinweis: </b>ContentContainer-Objekte mit Attachments, die durch Entschlsselung
   * von EncryptedDataOSCI-Objekten gewonnen wurden, knnen hier nicht ohne weiteres
   * hinzugefgt werden. In diesem Fall mssen die Attachments zunchst ausgelesen und
   * die Daten in neu angelegten Attachment-Objekten wieder hinzugefgt werden.
   * @param container Inhaltsdatencontainer
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public void addContentContainer(ContentContainer container)
                           throws OSCIRoleException
  {
    super.addContentContainer(container);
  }

  /**
   * Entfernt einen Inhaltsdatencontainer aus der Nachricht.
   * @param container Inhaltsdatencontainer
   * @see #addContentContainer
   */
  public void removeContentContainer(ContentContainer container)
  {
    super.removeContentContainer(container);
  }

  /**
   * Liefert die in die Nachricht eingestellten (unverschlsselten) Inhaltsdaten als ContentContainer-Objekte.
   * @return enthaltene ContentContainer mit Inhaltsdaten
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public ContentContainer[] getContentContainer()
  {
    return super.getContentContainer();
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> Inhaltsdaten nach dem ContentContainer
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen ContentContainer oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public ContentContainer getContentContainerByRefID(String refID)
  {
	  return super.getContentContainerByRefID(refID);
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> ContentContainer nach dem Content
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen Content oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public Content getContentByRefID(String refID)
  {
	return super.getContentByRefID(refID);
  }

  /**
   * Fgt der Nachricht einen EncryptedData-Eintrag mit verschlsselten
   * Inhaltsdaten hinzu.
   * @param encData verschlsselte Daten
   * @see EncryptedDataOSCI
   */
  public void addEncryptedData(EncryptedDataOSCI encData)
                        throws OSCIRoleException
  {
    super.addEncryptedData(encData);
  }

  /**
   * Entfernt einen EncryptedData-Eintrag mit verschlsselten Daten
   * aus der Nachricht.
   * @param encData verschlsselte Daten
   * @see #addEncryptedData
   * @see EncryptedDataOSCI
   */
  public void removeEncryptedData(EncryptedDataOSCI encData)
  {
    super.removeEncryptedData(encData);
  }

  /**
   * Liefert die in die Nachricht eingestellten verschlsselten Inhaltsdaten
   * als EncryptedData-Objekte.
   * @return enthaltene EncryptedData-Objekt mit verschlsselten Inhaltsdaten
   * @see EncryptedDataOSCI
   */
  public EncryptedDataOSCI[] getEncryptedData()
  {
    return super.getEncryptedData();
  }

  /**
   * Setzt die gewnschte Qualitt des Zeitstempels, mit dem der Intermedir
   * den Eingang des Auftrags im Laufzettel protokolliert.
   *
   * @param cryptographic <b>true</b>: kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b>: Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs, default)
   * @see #getQualityOfTimeStampCreation()
   */
  public void setQualityOfTimeStampCreation(boolean cryptographic)
  {
    qualityOfTimestampTypeCreation = new QualityOfTimestampH(false, cryptographic);
  }

  /**
   * Liefert die Qualitt des Zeitstempels, mit dem der Intermedir den
   * Eingang des Auftrags im Laufzettel protokolliert.
   * @return Qualitt des Zeitstempels: <b>true</b> - kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b> - Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs)
   * @see #setQualityOfTimeStampCreation(boolean)
   */
  public boolean getQualityOfTimeStampCreation()
  {
    return qualityOfTimestampTypeCreation.isQualityCryptographic();
  }

  /**
   * Setzt die gewnschte Qualitt des Zeitstempels, mit dem der Intermedir den
   * Empfang der Bearbeitungsantwort im Laufzettel protokolliert.
   * @param cryptographic <b>true</b>: kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b>: Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs, default)
   * @see #getQualityOfTimeStampReception()
   */
  public void setQualityOfTimeStampReception(boolean cryptographic)
  {
    qualityOfTimestampTypeReception = new QualityOfTimestampH(true, cryptographic);
  }

  /**
   * Liefert die geforderte Qualitt des Zeitstempels, mit dem der Intermedir den
   * Empfang der Annahmeantwort im Laufzettel protokolliert.
   * @return Qualitt des Zeitstempels: <b>true</b> - kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b> - Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs)
   * @see #setQualityOfTimeStampReception(boolean)
   */
  public boolean getQualityOfTimeStampReception()
  {
    return qualityOfTimestampTypeReception.isQualityCryptographic();
  }

  /**
   * Liefert die Adresse des Empfngers der Inhaltsdaten.
   * @return Adresse (URI)
   */
  public String getContentReceiver()
  {
    return uriReceiver.toString();
  }

  /**
   * Setzt die Adresse des Empfngers der Inhaltsdaten.
   * @param uri Adresse
   * @throws java.net.URISyntaxException bei Syntaxfehlern
   */
  public void setContentReceiver(String uri) throws java.net.URISyntaxException
  {
    this.uriReceiver = new java.net.URI(uri);
  }

  /**
   * Liefert den im Laufzettel enthaltenen Betreff-Eintrag.
   * @return den Betreff der Zustellung
   */
  public String getSubject()
  {
    return subject;
  }

  /**
   * Setzt den Betreff-Eintrag im Laufzettel. Der bergebene Parameter-String
   * mu den Vorschriften fr den primitiven XML-Datentyp "string" entsprechen,
   * darf also z.B. keine XML-Steuerzeichen (<, > o..) enthalten.
   * @param  subject  der Betreff
   */
  public void setSubject(String subject)
  {
    this.subject = subject;
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * @return Message-ID
   */
  public String getMessageId()
  {
    return messageId;
  }

  /**
   * Setzt die Message-ID der Nachricht.
   */
  public void setMessageId(String messageId)
  {
    this.messageId = messageId;
  }

  void sign()
     throws IOException,
            OSCIException,
            de.osci.osci12.common.OSCICancelledException,
            java.security.NoSuchAlgorithmException
  {
    super.sign();
    // Der NonIntemedCertHeader liegt hier weiter unten
    messageParts.set(3, null);
    messageParts.set(7, nonIntermediaryCertificatesH);
  }

  /**
   * undocumented
   *
   * @throws OSCIRoleException undocumented
   * @throws IOException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void compose() throws OSCIException,
                                  IOException,
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(3, null);

    if (dialogHandler.getControlblock().getChallenge() == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": Challenge");

    if (dialogHandler.getControlblock().getResponse() == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": Response");

    if (dialogHandler.getControlblock().getConversationID() == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": Conversation-Id");

    if (dialogHandler.getControlblock().getSequenceNumber() == -1)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": SequenceNumber");

    importAllCertificates();
    createNonIntermediaryCertificatesH();

    if (uriReceiver == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": URI - contentReceiver");

    String head = "<" + osciNSPrefix + ":ContentReceiver URI=\"" + uriReceiver.toString() + "\"></" + osciNSPrefix +
                  ":ContentReceiver>";

    if (messageId != null)
    {
      head += ("<" + osciNSPrefix + ":MessageId>" +
      de.osci.helper.Base64.encode(messageId.getBytes(Constants.CHAR_ENCODING)) + "</" + osciNSPrefix + ":MessageId>");

      if (subject != null)
        head += ("<" + osciNSPrefix + ":Subject>" + subject + "</" + osciNSPrefix + ":Subject>");
    }
    else
    {
      qualityOfTimestampTypeCreation = null;
      qualityOfTimestampTypeReception = null;
    }

    messageParts.add(qualityOfTimestampTypeCreation);
    messageParts.add(qualityOfTimestampTypeReception);
    osciH = new OsciH(HeaderTags.mediateDelivery.getElementName(), head);
    messageParts.add(osciH);
    messageParts.add(nonIntermediaryCertificatesH);
    if(featureDescription!=null && dialogHandler.isSendFeatureDescription())
    {
      messageParts.add(featureDescription);
    }
    messageParts.addAll(customHeaders);
    body = new Body(getContentContainer(), getEncryptedData());
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
