package de.osci.osci12.messagetypes;


import java.io.IOException;
import java.security.NoSuchAlgorithmException;

import de.osci.osci12.OSCIException;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.common.OSCICancelledException;
import de.osci.osci12.messageparts.Content;
import de.osci.osci12.messageparts.ContentContainer;
import de.osci.osci12.messageparts.EncryptedDataOSCI;
import de.osci.osci12.messageparts.Inspection;
import de.osci.osci12.messageparts.ProcessCardBundle;
import de.osci.osci12.messageparts.Timestamp;


/**
 * <p>
 * Dieses Klasse reprsentiert die Antwort des Intermedirs auf einen Zustellungsabholauftrag. Clients
 * erhalten vom Intermedir eine Instanz dieser Klasse, die eine Rckmeldung ber den Erfolg der Operation
 * (getFeedback()) sowie ggf. die angeforderten verschlsselten und/oder unverschlsselten Inhaltsdaten
 * einschl. des zugehrigen Laufzettels enthlt.
 * </p>
 * <p>
 * Copyright  2017 Freie Hansestadt Bremen
 * </p>
 * <p>
 * 
 * </p>
 * <p>
 * Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr freie Softwarebibliotheken genutzt
 * werden.
 * </p>
 * Die Lizenzbestimmungen knnen unter der URL <a href="<a href='www.osci.de' >www.osci.de</a >">www.osci.de</a> abgerufen oder bei der
 * Koordinierungsstelle fr IT-Standards (KoSIT), Schillerstr. 1, 28195 Bremen schriftlich angefordert werden.
 * </p>
 *
 * @author R. Lindemann
 * @version 1.8.0
 * @since 1.8.0
 * @see de.osci.osci12.messagetypes.FetchDelivery
 */
public abstract class ResponseToFetchAbstract extends OSCIResponseTo implements ContentPackageI
{

  // private static Log log = LogFactory.getLog(ResponseToFetchDelivery.class);
  String selectionRule;

  int selectionMode = -1;

  protected ProcessCardBundle processCardBundle;

  ResponseToFetchAbstract(DialogHandler dh)
  {
    super(dh);
  }

  /**
   * Liefert die in die Nachricht eingestellten (unverschlsselten) Inhaltsdaten als ContentContainer-Objekte.
   * 
   * @return enthaltene ContentContainer mit Inhaltsdaten
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public ContentContainer[] getContentContainer()
  {
    return super.getContentContainer();
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> Inhaltsdaten nach dem ContentContainer mit der bergebenen RefID.
   * 
   * @param refID zu suchende RefID
   * @return den zugehrigen ContentContainer oder null, wenn die Referenz nicht gefunden wurde.
   */
  public ContentContainer getContentContainerByRefID(String refID)
  {
    return super.getContentContainerByRefID(refID);
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> ContentContainer nach dem Content mit der bergebenen RefID.
   * 
   * @param refID zu suchende RefID
   * @return den zugehrigen Content oder null, wenn die Referenz nicht gefunden wurde.
   */
  public Content getContentByRefID(String refID)
  {
    return super.getContentByRefID(refID);
  }

  /**
   * Liefert die in die Nachricht eingestellten verschlsselten Inhaltsdaten als EncryptedData-Objekte.
   * 
   * @return enthaltene EncryptedData-Objekt mit verschlsselten Inhaltsdaten
   * @see EncryptedDataOSCI
   */
  public EncryptedDataOSCI[] getEncryptedData()
  {
    return super.getEncryptedData();
  }

  void setSelectionRule(String selectionRule)
  {
    this.selectionRule = selectionRule;
  }

  /**
   * Liefert die gesetzte Auswahlregel. Der Inhalt des zurckgegebenen Strings hngt vom gesetzten
   * Auswahlmodus ab und kann entweder in einer Base64-codierten Message-ID oder einem Datum bestehen. Das
   * Format eines Datums entspricht dem XML-Schema nach <a href="http://www.w3.org/TR/xmlschema-2/#dateTime">
   * http://www.w3.org/TR/xmlschema-2/#dateTime</a>. Wurde keine Regel gesetzt, wird als default null
   * zurckgegeben.
   * 
   * @return Auswahlregel (Message-ID oder Datum)
   * @see #getSelectionRule()
   */
  public String getSelectionRule()
  {
    return selectionRule;
  }

  void setSelectionMode(int selectionMode)
  {
    this.selectionMode = selectionMode;
  }

  /**
   * Liefert den gesetzten Auswahlmodus.
   * 
   * @return den Auswahlmodus SELECT_BY_MESSAGE_ID, SELECT_BY_DATE_OF_RECEPTION oder NO_SELECTION_RULE
   */
  public int getSelectionMode()
  {
    return selectionMode;
  }

  /**
   * Diese Methode liefert den Laufzettel der Zustellung zurck oder null, wenn bei der Verarbeitung der
   * Nachricht ein Fehler aufgetreten ist. Die Informationen im Laufzettel knnen auch direkt ber die
   * einzelnen getX()-Methoden ausgewertet werden.
   *
   * @return den Laufzettel als ProcessCardBundle-Objekt, im Fehlerfall null
   * @see #getTimestampCreation()
   * @see #getTimestampForwarding()
   * @see #getInspections()
   * @see #getSubject()
   * @see #getRecentModification()
   * @see #getMessageId()
   */
  public ProcessCardBundle getProcessCardBundle()
  {
    return processCardBundle;
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des Eingangs des Zustellungsauftrags beim
   * Intermedir.
   * 
   * @return Zeitstempel der Einreichung beim Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampCreation()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getCreation();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des vollstndigen Aufbaus der
   * Abholantwort vom Intermedir fr den Empfnger.
   * 
   * @return Zeitstempel der Erstellung der Abholantwort durch den Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampForwarding()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getForwarding();
  }

  /**
   * Liefert die Ergebnisse der Zertifikatsprfungen in Form von Inspection-Objekten, die im
   * ProcessCardBundle-Objekt enthalten sind.
   * 
   * @return inspections die Prfergebnisse
   * @see #getProcessCardBundle()
   */
  public Inspection[] getInspections()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getInspections();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Betreff-Eintrag.
   * 
   * @return den Betreff der Zustellung
   * @see #getProcessCardBundle()
   */
  public String getSubject()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getSubject();
  }

  /**
   * Liefert das Datum der letzten nderung des Laufzettels. Das Format entspricht dem XML-Schema nach
   * http://www.w3.org/TR/xmlschema-2/#dateTime
   * 
   * @return Datum der letzten nderung
   * @see #getProcessCardBundle()
   */
  public String getRecentModification()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getRecentModification();
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * 
   * @return Message-ID
   */
  public String getMessageId()
  {
    if (processCardBundle == null)
      return null;

    return processCardBundle.getMessageId();
  }

  void sign() throws IOException, OSCIException, OSCICancelledException, NoSuchAlgorithmException
  {
    super.sign();
    // Der IntemedCertHeader liegt hier weiter unten
    messageParts.set(4, intermediaryCertificatesH);
    messageParts.set(2, null);
  }


}
