package de.osci.osci12.messagetypes;

import java.io.IOException;
import java.io.OutputStream;
import java.security.NoSuchAlgorithmException;
import java.security.cert.X509Certificate;
import java.util.Vector;

import de.osci.helper.Base64;
import de.osci.helper.StoreOutputStream;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.Constants.HeaderTags;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.common.OSCICancelledException;
import de.osci.osci12.extinterfaces.crypto.Decrypter;
import de.osci.osci12.extinterfaces.crypto.Signer;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.messageparts.Content;
import de.osci.osci12.messageparts.ContentContainer;
import de.osci.osci12.messageparts.EncryptedDataOSCI;
import de.osci.osci12.roles.Addressee;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.roles.Originator;
import de.osci.osci12.soapheader.OsciH;
import de.osci.osci12.soapheader.QualityOfTimestampH;


/**
 * <p>Mit dieser Klasse werden Nachrichtenobjekte fr Bearbeitungsantworten
 * angelegt. Ein passiver Client, der als Supplier fungiert, muss nach Empfang
 * eines Bearbeitungsauftrags eine Instanz dieser Klasse aufbauen und an den Intermedir
 * zurcksenden. Die Nachricht enthlt eine Rckmeldung ber
 * den Empfang der Nachricht (Feedback) sowie ggf. verschlsselte bzw.
 * unverschlsselte Inhaltsdaten.</p>
 *
 * <p>Copyright  2017 Freie Hansestadt Bremen</p>
 * <p>Erstellt von Governikus GmbH & Co.  KG
 * </p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL <a href='www.osci.de' >www.osci.de</a > abgerufen oder bei der
 * Koordinierungsstelle fr IT-Standards (KoSIT),  Schillerstr. 1,
 * 28195 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.8.0
 *
 * @see de.osci.osci12.messagetypes.ProcessDelivery
 */
public class ResponseToProcessDelivery extends OSCIResponseTo implements ContentPackageI
{
  //  private static Log log = LogFactory.getLog(ResponseToProcessDelivery.class);
  private String subject;

  ResponseToProcessDelivery(DialogHandler dh)
  {
    super(dh);
    messageType = RESPONSE_TO_PROCESS_DELIVERY;
  }

  /**
   * Legt ein Nachrichtenobjekt fr eine Bearbeitungsantwort an.
   *
   * @see de.osci.osci12.common.DialogHandler
   */
  public ResponseToProcessDelivery(ProcessDelivery procDel)
  {
    super(procDel.dialogHandler);
    messageType = RESPONSE_TO_PROCESS_DELIVERY;

    // Nun wirds interessant: Weil in diesem Szenario die Rollen wechseln (der Orginator der Anfrage wird zum Addressee der Antwort
    // und umgekehrt) werden die Rollenobjekte umgebaut. Damit passen die Rollenobjekte des DialogHandlers nicht mehr so richtig,
    // aber in einem impliziten Dialog kann man das wohl riskieren.
    Signer signer = null;
    Decrypter decrypter = null;

    try
    {
      if (dialogHandler.getSupplier().hasSignaturePrivateKey())
        signer = dialogHandler.getSupplier().getSigner();

      if (dialogHandler.getSupplier().hasCipherPrivateKey())
        decrypter = dialogHandler.getSupplier().getDecrypter();

      originator = new Originator(signer, decrypter);

      X509Certificate signerCert = null;
      X509Certificate cipherCert = null;

      if (procDel.getOriginator().hasSignatureCertificate())
        signerCert = procDel.getOriginator().getSignatureCertificate();

      if (procDel.getOriginator().hasCipherCertificate())
        cipherCert = procDel.getOriginator().getCipherCertificate();

      addressee = new Addressee(signerCert, cipherCert);
      dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
      dialogHandler.getControlblock().setChallenge(null);
      dialogHandler.getControlblock().setConversationID(null);
      dialogHandler.getControlblock().setSequenceNumber(-1);
      setQualityOfTimeStampCreation(false);
      setQualityOfTimeStampReception(false);
    }
    catch (OSCIRoleException ex)
    {
      /* Nicht erforderlich */
    }
  }

  /**
   * Liefert den im Laufzettel enthaltenen Betreff-Eintrag.
   * @return den Betreff der Zustellung
   */
  public String getSubject()
  {
    return subject;
  }

  /**
   * Setzt den Betreff-Eintrag der Nachricht. Der bergebene Parameter-String
   * mu den Vorschriften fr den primitiven XML-Datentyp "string" entsprechen,
   * darf also z.B. keine XML-Steuerzeichen (<, > o..) enthalten.
   * @param  subject  der Betreff
   */
  public void setSubject(String subject)
  {
    this.subject = subject;
  }

  /**
   * Setzt die Rckmeldungen (Fehler und Warnungen) auf Auftragsebene.
   * @param code Array mit Fehlercodes
   */
  public void setFeedback(String[] code)
  {
    super.setFeedback(code);
  }

  /**
   * Setzt die Rckmeldungen (Feedback-Eintrge) mit frei whlbaren Texten und Sprachen.
   * Der erste Index des Arrays entspricht dem Index des Entry-Elementes.
   * Beim zweiten Index bezeichnet <br>
   * 0 - das Sprachkrzel (z.B. "de", "en-US")<br>
   * 1 - den Code<br>
   * 2 - den Text<br>
   */
  public void setFeedback(String[][] entries)
  {
    feedBack = new Vector<String[]>();
    for (int i = 0; i < entries.length; i++)
      feedBack.add(entries[i]);
  }


  /**
   * Setzt die gewnschte Qualitt des Zeitstempels, mit dem der Intermedir
   * den Eingang des Auftrags im Laufzettel protokolliert.
   *
   * @param cryptographic <b>true</b>: kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b>: Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs, default)
   * @see #getQualityOfTimeStampCreation()
   */
  public void setQualityOfTimeStampCreation(boolean cryptographic)
  {
    qualityOfTimestampTypeCreation = new QualityOfTimestampH(false, cryptographic);
  }

  /**
   * Liefert die Qualitt des Zeitstempels, mit dem der Intermedir den
   * Eingang des Auftrags im Laufzettel protokolliert.
   * @return Qualitt des Zeitstempels: <b>true</b> - kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b> - Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs)
   * @see #setQualityOfTimeStampCreation(boolean)
   */
  public boolean getQualityOfTimeStampCreation()
  {
    return qualityOfTimestampTypeCreation.isQualityCryptographic();
  }

  /**
   * Setzt die gewnschte Qualitt des Zeitstempels, mit dem der Intermedir die
   * Empfangsbesttigung der Zustellung durch den Empfnger im Laufzettel protokolliert.
   * Die Empfangsbesttigung besteht in einem weiteren Auftrag, den der Empfnger
   * nach Erhalt der Bearbeitungsantwort innerhalb desselben expliziten Dialogs
   * an den Intermedir schickt.
   * @param cryptographic <b>true</b>: kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b>: Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs)
   * @see #getQualityOfTimeStampReception()
   */
  public void setQualityOfTimeStampReception(boolean cryptographic)
  {
    qualityOfTimestampTypeReception = new QualityOfTimestampH(true, cryptographic);
  }

  /**
   * Liefert die geforderte Qualitt des Zeitstempels, mit dem der Intermedir den
   * Empfang der Annahmeantwort im Laufzettel protokolliert.
   * @return Qualitt des Zeitstempels: <b>true</b> - kryptographischer Zeitstempel von einem
   * akkreditierten Zeitstempeldienst<br><b>false</b> - Einfacher Zeitstempel
   * (lokale Rechnerzeit des Intermedirs, default)
   * @see #setQualityOfTimeStampReception(boolean)
   */
  public boolean getQualityOfTimeStampReception()
  {
    return qualityOfTimestampTypeReception.isQualityCryptographic();
  }

  /**
   * Fgt der Nachricht einen Inhaltsdatencontainer hinzu. Diese Methode
   * sollte erst aufgerufen werden, wenn der Container vollstndig erstellt wurde.
   * <b>Hinweis: </b>ContentContainer-Objekte mit Attachments, die durch Entschlsselung
   * von EncryptedDataOSCI-Objekten gewonnen wurden, knnen hier nicht ohne weiteres
   * hinzugefgt werden. In diesem Fall mssen die Attachments zunchst ausgelesen und
   * die Daten in neu angelegten Attachment-Objekten wieder hinzugefgt werden.
   * @param container Inhaltsdatencontainer
   * @see de.osci.osci12.messageparts.ContentContainer
   */
  public void addContentContainer(ContentContainer container)
                           throws OSCIRoleException
  {
    super.addContentContainer(container);
  }

  /**
   * Entfernt einen Inhaltsdatencontainer aus der Nachricht.
   * @param container Inhaltsdatencontainer
   * @see #addContentContainer
   */
  public void removeContentContainer(ContentContainer container)
  {
    super.removeContentContainer(container);
  }

  /**
   * Liefert ein Array mit den Inhaltsdatencontainern der Nachricht.
   * @see #addContentContainer
   */
  public ContentContainer[] getContentContainer()
  {
    return super.getContentContainer();
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> Inhaltsdaten nach dem ContentContainer
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen ContentContainer oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public ContentContainer getContentContainerByRefID(String refID)
  {
	  return super.getContentContainerByRefID(refID);
  }

  /**
   * Durchsucht <b>die unverschlsselten</b> ContentContainer nach dem Content
   * mit der bergebenen RefID.
   * @param refID zu suchende RefID
   * @return den zugehrigen Content oder null, wenn die Referenz
   * nicht gefunden wurde.
   */
  public Content getContentByRefID(String refID)
  {
	return super.getContentByRefID(refID);
  }

  /**
   * Fgt der Nachricht einen EncryptedData-Eintrag mit verschlsselten
   * Inhaltsdaten hinzu.
   * @param encData verschlsselte Daten
   * @see EncryptedDataOSCI
   */
  public void addEncryptedData(EncryptedDataOSCI encData)
                        throws OSCIRoleException
  {
    super.addEncryptedData(encData);
  }

  /**
   * Entfernt einen EncryptedData-Eintrag mit verschlsselten Daten
   * aus der Nachricht.
   * @param encData verschlsselte Daten
   * @see #addEncryptedData
   * @see EncryptedDataOSCI
   */
  public void removeEncryptedData(EncryptedDataOSCI encData)
  {
    super.removeEncryptedData(encData);
  }

  /**
   * Liefert die in die Nachricht eingestellten verschlsselten Inhaltsdaten
   * als EncryptedData-Objekte.
   * @return enthaltene EncryptedData-Objekt mit verschlsselten Inhaltsdaten
   * @see EncryptedDataOSCI
   */
  public EncryptedDataOSCI[] getEncryptedData()
  {
    return super.getEncryptedData();
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * @return Message-ID
   */
  public String getMessageId()
  {
    return messageId;
  }

  /**
   * Setzt die Message-ID der Nachricht.
   */
  public void setMessageId(String messageId)
  {
    this.messageId = messageId;
  }

  /**
   * Bringt eine Supplier-Signatur an.
   * @throws IOException bei Schreib-/Leseproblemen
   * @throws OSCIRoleException wenn dem Rollenobjekt, das als Client fungiert,
   * kein Signer-Objekt zugeordnet wurde
   * @throws OSCICancelledException bei Abbruch durch den
   * @throws NoSuchAlgorithmException
   * Benutzer
   */
  public void sign()
            throws IOException,
                   OSCIException,
                   OSCICancelledException,
                   NoSuchAlgorithmException
  {
    super.sign(originator);
    //    messageParts.set(1, signatureHeader);
    messageParts.set(2, null);
    messageParts.set(6, nonIntermediaryCertificatesH);
  }

  /**
   * undocumented
   *
   * @throws OSCIRoleException undocumented
   * @throws IOException undocumented
   * @throws NoSuchAlgorithmException
   */
  protected void compose() throws OSCIException,
                                  IOException,
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(2, null);

    String head = writeFeedBack();

    if (messageId != null)
    {
      head += ("<" + osciNSPrefix + ":MessageId>" +
      Base64.encode(messageId.getBytes(Constants.CHAR_ENCODING)) + "</" + osciNSPrefix + ":MessageId>");

      if (subject != null)
        head += ("<" + osciNSPrefix + ":Subject>" + subject + "</" + osciNSPrefix + ":Subject>");
    }
    else
    {
      qualityOfTimestampTypeCreation = null;
      qualityOfTimestampTypeReception = null;
    }

    messageParts.add(qualityOfTimestampTypeCreation);
    messageParts.add(qualityOfTimestampTypeReception);
    osciH = new OsciH(HeaderTags.responseToProcessDelivery.getElementName(), head);
    messageParts.add(osciH);
    createNonIntermediaryCertificatesH();
    messageParts.add(nonIntermediaryCertificatesH);
    body = new Body(getContentContainer(), getEncryptedData());
    if(featureDescription!=null && dialogHandler.isSendFeatureDescription())
    {
      messageParts.add(featureDescription);
    }
    messageParts.addAll(customHeaders);
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }

  /**
   * Serialisiert und schreibt die Nachricht - ggf. verschlsselt - in den
   * bergebenen Stream. Es wird abhngig vom isEncryption()-Flag des
   * DialogHandlers verschlsselt oder nicht, Signaturen mssen jedoch von der
   * Anwendung selbst vor Aufruf dieser Methode (sign()) angebracht werden.
   * Die ausgehende Nachricht kann zu Debug- oder Archivierungszwecken (in jedem Fall
   * unverschlsselt) in den zweiten bergebenen Stream geschrieben werden.
   * Dieser Parameter kann null sein.
   * @param out Stream, in den die Antwortnachricht geschrieben werden soll
   * @param storeOutput Stream, in dem die (unverschlsselte) Antwortnachricht
   * gespeichert werden soll
   * @throws OSCIRoleException wenn erforderliche Zertifikate fehlen
   * @throws IOException bei Schreibproblemen
   * @throws NoSuchAlgorithmException wenn ein bentigter
   * Algorithmus nicht untersttzt wird
   * @see #sign()
   */
  public void writeToStream(OutputStream out, OutputStream storeOutput)
                     throws IOException,
                            OSCIException,
                            NoSuchAlgorithmException
  {
    if (dialogHandler.isEncryption())
      new SOAPMessageEncrypted(this, storeOutput).writeXML(out);
    else if (storeOutput != null)
    {
      StoreOutputStream sos = new StoreOutputStream(out, storeOutput);
      writeXML(sos);
      sos.close();
    }
    else
      writeXML(out);
  }
}
