package de.osci.osci12.messagetypes;

import java.io.IOException;
import java.security.NoSuchAlgorithmException;

import de.osci.helper.Tools;
import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants.HeaderTags;
import de.osci.osci12.common.Constants.LanguageTextEntries;
import de.osci.osci12.common.DialogHandler;
import de.osci.osci12.common.OSCICancelledException;
import de.osci.osci12.messageparts.Body;
import de.osci.osci12.messageparts.Inspection;
import de.osci.osci12.messageparts.ProcessCardBundle;
import de.osci.osci12.messageparts.Timestamp;
import de.osci.osci12.roles.OSCIRoleException;
import de.osci.osci12.roles.Originator;
import de.osci.osci12.soapheader.OsciH;


/**
 * <p>Instanzen dieser Klasse werden als Antworten auf Zustellungsauftrge
 * zurckgegeben. Das Nachrichtenobjekt enthlt eine Rckmeldung ber den Erfolg
 * der Operation (getFeedback()) sowie ggf. den Laufzettel der
 * Zustellung.</p>
 *
 * <p>Copyright  2017 Freie Hansestadt Bremen</p>
 * <p>Erstellt von Governikus GmbH & Co.  KG
 * </p>
 * <p>Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr
 * freie Softwarebibliotheken genutzt werden.</p>Die Lizenzbestimmungen knnen
 * unter der URL <a href='www.osci.de' >www.osci.de</a > abgerufen oder bei der
 * Koordinierungsstelle fr IT-Standards (KoSIT),  Schillerstr. 1,
 * 28195 Bremen schriftlich angefordert werden.</p>
 *
 *
 * @author R. Lindemann, N. Bngener
 * @version 1.8.0
 *
 * @see de.osci.osci12.messagetypes.StoreDelivery
 * @see #getProcessCardBundle()
 */
public class ResponseToStoreDelivery extends OSCIResponseTo
{
  //  private static Log log = LogFactory.getLog(ResponseToStoreDelivery.class);
  ProcessCardBundle processCardBundle;

  ResponseToStoreDelivery(DialogHandler dh) throws Exception
  {
    this(dh, false);
  }

  ResponseToStoreDelivery(DialogHandler dh, boolean parser)
                   throws NoSuchAlgorithmException
  {
    super(dh);
    messageType = RESPONSE_TO_STORE_DELIVERY;
    originator = ((Originator) dh.getClient());

    if (!parser)
    {
      dialogHandler.getControlblock().setResponse(dialogHandler.prevChallenge);
      dialogHandler.getControlblock().setChallenge(Tools.createRandom(10));
    }

    body = new Body("");
    body.setNSPrefixes(this);
  }

  /**
   * Diese Methode liefert den Laufzettel der Zustellung zurck oder null,
   * wenn bei der Verarbeitung der Nachricht ein Fehler aufgetreten ist.
   * Die Informationen im Laufzettel knnen auch direkt ber die einzelnen
   * getX()-Methoden ausgewertet werden.
   *
   * @return den Laufzettel als ProcessCardBundle-Objekt, im Fehlerfall null
   * @see #getTimestampCreation()
   * @see #getInspections()
   * @see #getSubject()
   * @see #getRecentModification()
   * @see #getMessageId()
   */
  public ProcessCardBundle getProcessCardBundle()
  {
    return processCardBundle;
  }

  /**
   * Liefert den im Laufzettel enthaltenen Zeitstempel vom Zeitpunkt des Eingangs
   * des Zustellungsauftrags beim Intermedir.
   * @return Zeitstempel der Einreichung beim Intermedir
   * @see #getProcessCardBundle()
   */
  public Timestamp getTimestampCreation()
  {
    return processCardBundle.getCreation();
  }

  /**
   * Liefert die Ergebnisse der Zertifikatsprfungen in Form von Inspection-Objekten,
   * die im ProcessCardBundle-Objekt enthalten sind.
   * @return inspections die Prfergebnisse
   * @see #getProcessCardBundle()
   *
   */
  public Inspection[] getInspections()
  {
    return processCardBundle.getInspections();
  }

  /**
   * Liefert den im Laufzettel enthaltenen Betreff-Eintrag.
   * @return den Betreff der Zustellung
   * @see #getProcessCardBundle()
   */
  public String getSubject()
  {
    return processCardBundle.getSubject();
  }

  /**
   * Liefert das Datum der letzten nderung des Laufzettels. Das Format
   * entspricht dem XML-Schema nach http://www.w3.org/TR/xmlschema-2/#dateTime
   * @return Datum der letzten nderung
   * @see #getProcessCardBundle()
   */
  public String getRecentModification()
  {
    return processCardBundle.getRecentModification();
  }

  /**
   * Liefert die Message-ID der Nachricht.
   * @return Message-ID
   * @see #getProcessCardBundle()
   */
  public String getMessageId()
  {
    return processCardBundle.getMessageId();
  }

  void sign()
     throws IOException, 
            OSCIException, 
            OSCICancelledException, 
            NoSuchAlgorithmException
  {
    super.sign();
    // Der IntemedCertHeader liegt hier weiter unten
    messageParts.set(4, intermediaryCertificatesH);
    messageParts.set(2, null);
  }

  /**
   * undocumented
   *
   * @throws IOException undocumented
   * @throws OSCIRoleException undocumented
   * @throws IllegalStateException undocumented
   */
  protected void compose() throws IOException, 
                                  OSCIException, 
                                  NoSuchAlgorithmException
  {
    super.compose();
    messageParts.set(2, null);

    if (dialogHandler.getControlblock().getResponse() == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": Response");

    if (dialogHandler.getControlblock().getConversationID() == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": Conversation-Id");

    if (dialogHandler.getControlblock().getSequenceNumber() == -1)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": SequenceNumber");

    if (feedBack == null)
      throw new IllegalStateException(DialogHandler.text.getString(LanguageTextEntries.missing_entry.name()) + ": Feedback");

    if (processCardBundle == null)
      osciH = new OsciH(HeaderTags.responseToStoreDelivery.getElementName(), writeFeedBack());
    else
      osciH = new OsciH(HeaderTags.responseToStoreDelivery.getElementName(), writeFeedBack() + processCardBundle.writeToString());

    messageParts.add(osciH);
    messageParts.add(intermediaryCertificatesH);
    if(featureDescription!=null && dialogHandler.isSendFeatureDescription())
    {
      messageParts.add(featureDescription);
    }
    messageParts.addAll(customHeaders);
    messageParts.add(body);
    stateOfMsg |= STATE_COMPOSED;
  }
}
