package de.osci.osci12.soapheader;

import java.io.IOException;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import de.osci.osci12.OSCIException;
import de.osci.osci12.common.Constants;
import de.osci.osci12.common.Constants.HeaderTags;
import de.osci.osci12.common.Constants.OSCIFeatures;


/**
 * <p>
 * Mit dieser Klasse wird der FeatureDescription Header bestckt. Die Chunk-Eigenschaften sowie weitere
 * Features knnen gesetzt werden. Dieser Header kann dann auf den Nachrichten-Objekten gesetzt werden.
 * </p>
 * <p>
 * Copyright  2017 Freie Hansestadt Bremen
 * </p>
 * <p>
 * 
 * </p>
 * <p>
 * Diese Bibliothek kann von jedermann nach Magabe der Bremer Lizenz fr freie Softwarebibliotheken genutzt
 * werden.
 * </p>
 * Die Lizenzbestimmungen knnen unter der URL <a href='<a href='www.osci.de' >www.osci.de</a >'>www.osci.de</a> abgerufen oder bei der
 * Koordinierungsstelle fr IT-Standards (KoSIT), Schillerstr. 1, 28195 Bremen schriftlich angefordert werden.
 * </p>
 *
 * @author R. Lindemann
 * @version 1.8.0
 * @since 1.8.0
 */
public class FeatureDescriptionH extends HeaderEntry
{

  private List<OSCIFeatures> supportedFeatures = new ArrayList<OSCIFeatures>();

  /*
   * Initial mit dem neutralem Wert -1 instanziert
   */
  private long maxMessageSize = -1;

  private long maxChunkSize = -1;

  private long minChunkSize = -1;

  private long chunkMessageTimeout = -1;

  /**
   * Dies sind die default Features der aktuellen Version
   */
  public static final FeatureDescriptionH defaultFeatureDescription = new FeatureDescriptionH();
  static
  {
    defaultFeatureDescription.getSupportedFeatures().add(OSCIFeatures.DisableBase64);
    defaultFeatureDescription.getSupportedFeatures().add(OSCIFeatures.GCMPaddingModus);
    defaultFeatureDescription.getSupportedFeatures().add(OSCIFeatures.OAEPEncryption);
    defaultFeatureDescription.getSupportedFeatures().add(OSCIFeatures.PartialMessageTransmission);
    defaultFeatureDescription.getSupportedFeatures().add(OSCIFeatures.SHA3HashAlgo);
  }


  /**
   * @return Liefert die Liste der eingetragenen untersttzten Features
   * @see OSCIFeatures
   * @see #setSupportedFeatures(List)
   */
  public List<OSCIFeatures> getSupportedFeatures()
  {
    return supportedFeatures;
  }

  /**
   * @param supportedFeatures Setzt die Liste der untersttzten Features. Wird zumeist vom
   *          Supplier (Intermedir) gesetzt.
   * @see OSCIFeatures
   * @see #getSupportedFeatures()
   */
  public void setSupportedFeatures(List<OSCIFeatures> supportedFeatures)
  {
    this.supportedFeatures = supportedFeatures;
  }

  /**
   * @return Liefert den eingestellten Wert der maximalen Nachrichtengre
   * @see #setMaxMessageSize(long)
   */
  public long getMaxMessageSize()
  {
    return maxMessageSize;
  }

  /**
   * @param maxMessageSize Setzt den Wert der maximalen Nachrichtengre. Wird zumeist vom
   *          Supplier (Intermedir) gesetzt.
   * @see #getMaxMessageSize()
   */
  public void setMaxMessageSize(long maxMessageSize)
  {
    this.maxMessageSize = maxMessageSize;
  }


  /**
   * @return Liefert die maximale akzeptierte Chunk Gre
   * @see #setMaxChunkSize(long)
   */
  public long getMaxChunkSize()
  {
    return maxChunkSize;
  }

  /**
   * @param maxChunkSize Setzt die maximale Chunk Gre. Wird zumeist vom Supplier (Intermedir) gesetzt.
   * @see #getMaxChunkSize()
   */
  public void setMaxChunkSize(long maxChunkSize)
  {
    this.maxChunkSize = maxChunkSize;
  }

  /**
   * @return Liefert die minimale Chunk Gre
   * @see #setMinChunkSize(long)
   */
  public long getMinChunkSize()
  {
    return minChunkSize;
  }

  /**
   * @param minChunkSize Setzt die minimale Chunk Gre. Wird zumeist vom Supplier (Intermedir) gesetzt.
   * @see #getMinChunkSize()
   */
  public void setMinChunkSize(long minChunkSize)
  {
    this.minChunkSize = minChunkSize;
  }

  /**
   * @return Liefert die garantierte Timeout-Zeit vom Supplier, nach der eine Rckantwort generiert wird. Sollte
   *         die Verarbeitung beim Supplier das Timeout berschreiten, wird eine vorlufige Rckantwort
   *         generiert.
   * @see #setChunkMessageTimeout(long)
   */
  public long getChunkMessageTimeout()
  {
    return chunkMessageTimeout;
  }

  /**
   * @param chunkMessageTimeout Setzt die Timeout-Zeit. Wird zumeist vom Supplier (Intermedir) gesetzt.
   * @see #getChunkMessageTimeout()
   */
  public void setChunkMessageTimeout(long chunkMessageTimeout)
  {
    this.chunkMessageTimeout = chunkMessageTimeout;
  }

  @Override
  protected void writeXML(OutputStream out) throws IOException, OSCIException
  {
    String namespacePre = osci2017NSPrefix;
    out.write(("<" + namespacePre + ":"
               + HeaderTags.FeatureDescription.getElementName()).getBytes(Constants.CHAR_ENCODING));
    out.write(ns2017);

    // lexikografische Reihenfolge der Attribute ist entscheidend fuer die Signaturpruefung!

    if (chunkMessageTimeout != -1)
      out.write((" ChunkMessageTimeout=\"" + chunkMessageTimeout + "\"").getBytes(Constants.CHAR_ENCODING));

    out.write((" Id=\"" + id + "\"").getBytes(Constants.CHAR_ENCODING));

    if (maxChunkSize != -1)
      out.write((" MaxChunkSize=\"" + maxChunkSize + "\"").getBytes(Constants.CHAR_ENCODING));


    if (maxMessageSize != -1)
      out.write((" MaxMessageSize=\"" + maxMessageSize + "\"").getBytes(Constants.CHAR_ENCODING));

    if (minChunkSize != -1)
      out.write((" MinChunkSize=\"" + minChunkSize + "\"").getBytes(Constants.CHAR_ENCODING));


    out.write(">".getBytes());
    if (supportedFeatures != null && !supportedFeatures.isEmpty())
    {
      StringBuilder sb = new StringBuilder("<");
      sb.append(namespacePre);
      sb.append(":SupportedFeatures>");

      for ( OSCIFeatures feature : supportedFeatures )
      {
        sb.append("<");
        sb.append(namespacePre);
        sb.append(":Feature Key=\"" + feature.name() + "\"");
        sb.append(" Version=\"" + feature.getVersion() + "\">");
        sb.append("</");
        sb.append(namespacePre);
        sb.append(":Feature>");
      }
      sb.append("</");
      sb.append(namespacePre);
      sb.append(":SupportedFeatures>");
      out.write(sb.toString().getBytes());
    }
    out.write(("</" + namespacePre + ":" + HeaderTags.FeatureDescription.getElementName()
               + ">").getBytes(Constants.CHAR_ENCODING));
  }

}
